package Data::Filter;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Data::Filter - Class performing object filtering

=head1 SYNOPSIS

 use Data::Filter;

 # During the instantination of the filter pass the name of the file containing the filter grammmar
 # and name of the file containing filtering rules as arguments
 my $filter = Data::Filter->new(filter_grammar_file => $filter_grammar_file,
                                filter_config => $filter_rules,
                                dt_resolver => $resolver);

=head1 DESCRIPTION

Implemented filtering algorithm is inspired by Apache webserver. There are two
chains (or groups) of filtering rules, 'allow' and 'deny'. Rules in 'allow' chan describe what
objects may pass, 'deny' chain rules describe objects, that will be rejected. Each chain
may have multiple filter rules defined and they are processed in order of appearance
in config file and by default logically ORed together. If the object matches a 'allow' rule
in the chain, it is immediatelly accepted and further processing is skipped. If
the message matches a 'deny' rule in the chain, it is immediatelly rejected and
further processing is skipped. If the message did not matched any rule in any chain,
the result will be the defined default policy ('allow' or 'deny'). The order of
processing of the chains as a whole, if the 'allow' or 'deny' rules will be processed
first, is configurable option.

=head1 FILTER RULES CONFIGURATION FILE SYNTAX

Configuration file uses the JSON format. All configurations must be defined as attributes
of parent object. Following example describes all available configuration options:

 {
    # Order in which to process the filter chains
    # [DEFAULT='allow']
    "order": "allow",
    #"order": "allow, deny",
    #"order": "allow",
    #"order": "deny, allow",
    #"order": "deny",

    # Default action if the object did not match any rule in any filter chain
    # [DEFAULT='deny']
    "default_action": "deny",
    #"default_action": "allow",

    # Logical operation to be applied to rules in a 'allow' chain
    # [DEFAULT='or']
    "logop_allow": "or",
    #"logop_allow": "and",

    # Logical operation to be applied to rules in a 'deny' chain
    # [DEFAULT='or']
    "logop_deny": "or",
    #"logop_deny": "and",

    # Rules populating the 'allow' chain (processed in order of appearance)
    "allow": [
        "? _CESNET",
        "Description == \"Scan IPMI\" and Source/IP4 eq \"192.0.2.73\""
    ],

    # Rules populating the 'deny' chain (processed in order of appearance)
    "deny": [
        "? _TEST",
        "Description == \"Scan IPMI\" and Source/IP4 eq \"192.0.2.73\""
    ]
 }

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1), Data::Query::Object().

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Config::JSON;
use JSON::Schema::Validator;
use Data::Filter::Factory;
use Data::Filter::Compiler;
use Data::Convertor;
use Data::Query::Object;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

use constant CHAIN_ALLOW => 'allow';
use constant CHAIN_DENY  => 'deny';

use constant ACTION_ALLOW => 'allow';
use constant ACTION_DENY  => 'deny';

use constant ORDER_ALLOW => 'allow';
use constant ORDER_DENY  => 'deny';

use constant LOGOP_AND => 'and';
use constant LOGOP_OR  => 'or';

use constant DEFAULT_ORDER   => ORDER_ALLOW;
use constant DEFAULT_ACTION  => ACTION_DENY;
use constant DEFAULT_LOGOP   => LOGOP_OR;
use constant DEFAULT_COMPILE => '';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Hash containing all valid filter rule orders
my $VALID_ORDERS = {
        'deny,allow'    => ORDER_DENY,
        'allow,deny'    => ORDER_ALLOW,
        'deny, allow'   => ORDER_DENY,
        'allow, deny'   => ORDER_ALLOW,
        'deny'          => ORDER_DENY,
        'allow'         => ORDER_ALLOW,
    };

# Hash containing all valid filter chain logical operations
my $VALID_LOGOPS = {
        LOGOP_AND,  '_check_chain_anded',
        LOGOP_OR,   '_check_chain_ored',
    };

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

# Operator overloading
use overload '""' => 'to_string';

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new [PUBLIC,STATIC]

  Usage     : my $instance = Data::Filter->new()
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Croaks, if invoked on object
  Comments  : Internally uses _init() method
  See Also  : _init() method

=cut

sub new
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my $self = bless ({}, $class);
    return $self->_init(@_);
}

=item filter($) [PUBLIC]

  Usage     : my ($result, $filter, $index) = $instance->filter($object);
  Purpose   : Attempt to filter given object through configured filter structure
  Arguments : OBJECT OR HASH REFERENCE $object - filtered object
  Returns   : ARRAY OF
                BOOL $result - filtering result (0|1)
                STRING $filter - Applied filter
                INTEGER $index - Index of applied filter
  Throws    : Croaks, if invoked on class

=cut

sub filter($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($object,) = @_;
    confess "Filtered object must be blessed object or hash reference" unless ref $object eq 'HASH' or blessed($object);

    # Run the filter chains in speciffic order
    foreach my $chain (@{$self->{CHAINS}}) {
        # Determine the method name for actual filtering
        my $cmp_method_name = $chain->{'l'};

        # Check the filters in the chain in given mode
        my ($result, $rule, $idx) = $self->$cmp_method_name($chain->{'f'}, $chain->{'t'}, $object);

        # Defined result means authoritative answer (message either matched a rule
        # in 'allow' chain, or was denied by a rule in 'deny' chain)
        if (defined $result) {
            # Message denied
            if ($result == 0) {
                print STDERR "Result: message denied (rule: $idx)\n" if $DEVEL; #-+-> DEVEL ONLY <-+-#
                return (0, $rule, $idx);
            }
            # Message allowed
            elsif ($result == 1) {
                print STDERR "Result: message allowed (rule: $idx)\n" if $DEVEL; #-+-> DEVEL ONLY <-+-#
                return (1, $rule, $idx);
            }
        }
    }

    # Message did not matched any rule in any chain, return the default action
    if ($self->{DEFAULT_ACTION} eq ACTION_DENY) {
        print STDERR "Result: message denied by default policy\n" if $DEVEL; #-+-> DEVEL ONLY <-+-#
        return (0,'__DEFAULT__', 0);
    }
    elsif ($self->{DEFAULT_ACTION} eq ACTION_ALLOW) {
        print STDERR "Result: message allowed by default policy\n" if $DEVEL; #-+-> DEVEL ONLY <-+-#
        return (1,'__DEFAULT__', 0);
    }

    # We should never get to this point, but Mr. Murphy says otherwise
    die "Invalid default filter action '$self->{DEFAULT_ACTION}'\n";
}

=item to_string() [PUBLIC, ABSTRACT]

  Usage     : my $result = Data::Filter->to_string();
  Purpose   :
  Returns   :
  Arguments :
  Throws    :
  Comments  :
  See Also  :

=cut

sub to_string()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my $result = '';
    $result .= "Order: '".$self->{ORDER}."', default: '".$self->{DEFAULT_ACTION}."'\n";
    foreach my $ch (@{$self->{CHAINS}}) {
        $result .= "Chain: '".$ch->{t}."', logop: '".$self->{uc("logop_".$ch->{t})}."':\n";
        foreach my $f (@{$ch->{f}}) {
            $result .= " - '".$f->{f}."'\n";
        }
    }
    return $result;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init($$) [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Processor::Cog instance
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my %args = @_;
    confess "Filter grammar file must be given as argument" unless $args{filter_grammar_file};
    confess "Filter config must be given as argument" unless $args{filter_config};
    confess "Datatype resolver must be given as argument" unless $args{dt_resolver};

    # Load filter configurations from external file
    my $config = $self->_load_filter_config($args{filter_config}, $args{filter_schema_file});

    $self->{QUERY}          = Data::Query::Object->new(dt_resolver => $args{dt_resolver});
    $self->{FFACTORY}       = Data::Filter::Factory->new_from_file($args{filter_grammar_file});
    $self->{FCOMPILER}      = Data::Filter::Compiler->new(dt_resolver => $args{dt_resolver});
    $self->{ORDER}          = (exists $config->{'order'} and $VALID_ORDERS->{$config->{'order'}}) ? $VALID_ORDERS->{$config->{'order'}} : DEFAULT_ORDER;
    $self->{COMPILE}        = (exists $config->{'compile'})        ? $config->{'compile'}        : DEFAULT_COMPILE;
    $self->{DEFAULT_ACTION} = (exists $config->{'default_action'}) ? $config->{'default_action'} : DEFAULT_ACTION;
    $self->{LOGOP_ALLOW}    = (exists $config->{'logop_allow'})    ? $config->{'logop_allow'}    : DEFAULT_LOGOP;
    $self->{LOGOP_DENY}     = (exists $config->{'logop_deny'})     ? $config->{'logop_deny'}     : DEFAULT_LOGOP;

    # Determine the order of deny and allow chains
    my $allow_order = ($self->{ORDER} eq ORDER_ALLOW) ? 0 : 1;
    my $deny_order  = ($self->{ORDER} eq ORDER_DENY) ? 0 : 1;

    # Adjust the order if any of the filter chain was not defined
    $allow_order = $deny_order = 0 if not $config->{'allow'} or not $config->{'deny'};

    #die "No filtering rules were given\n" unless ($config->{'allow'} or $config->{'deny'});

    # Instantinate filter tree structures
    if ($config->{'allow'}) {
        $self->{CHAINS}->[$allow_order] = {
            't' => CHAIN_ALLOW,
            'f' => $self->_build_filters($config->{'allow'}),
            'l' => $VALID_LOGOPS->{$self->{LOGOP_ALLOW}},
        };
    }
    if ($config->{'deny'}) {
        $self->{CHAINS}->[$deny_order] = {
                't' => CHAIN_DENY,
                'f' => $self->_build_filters($config->{'deny'}),
                'l' => $VALID_LOGOPS->{$self->{LOGOP_DENY}},
            };
    }

    delete $self->{FFACTORY};
    return $self;
}

# _load_filter_config($;$) [PROTECTED]
#
#  Usage     : my $config = $self->_load_filter_config($filter_config, $filter_schema_file);
#  Purpose   : Load and prepare filter configurations
#  Arguments : STRING or HASH REFERENCE $filter_config - file name or bare configurations
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _load_filter_config($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter_config, $filter_schema_file) = @_;

    my $config = $filter_config;

    # If the configurations were not given bare
    my $error;
    unless (ref $filter_config eq 'HASH') {
        my $cfg_parser = Config::JSON->new($filter_schema_file);

        # Load the filter configuration from external file
        ($config, $error) = $cfg_parser->config($filter_config, $filter_schema_file);
        die "Filter configuration file '$filter_config' syntax error: '$error'\n" unless $config;
    }
    # Perform the validation, if available
    elsif ($filter_schema_file) {
        my $validator = JSON::Schema::Validator->new($filter_schema_file);
        $error = $validator->validate($config);
        die "Filter specification syntax error: '$error'\n" if $error;
    }

    return $config;
}

# _build_filters($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _build_filters($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter_rules,) = @_;
    confess "Filter rules must be given via array reference" unless $filter_rules and ref $filter_rules eq 'ARRAY';

    my $counter = 0;
    my $filters = [];
    foreach my $rule (@{$filter_rules}) {
        my $filter = $self->{FFACTORY}->filter_from_string($rule);
        $filter = $self->{FCOMPILER}->compile($filter, $self->{COMPILE}) if $self->{COMPILE};
        #push(@$filters, {'f'=>$filter,'r'=>$rule,'c'=>++$counter});
        push(@$filters, {'f'=>$filter,'r'=>"$filter",'c'=>++$counter});
    }
    return $filters;
}

# _check_chain_ored($$$) [PROTECTED]
#
#  Usage     :
#  Purpose   : Filter the given message through selected filter chain
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _check_chain_ored($$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter_chain, $chain_type, $alert_message) = @_;

    # Walk through all filters in selected chain
    foreach my $filter (@$filter_chain) {
        print STDERR "Checking filter '$filter->{'r'}' in chain '$chain_type'\n" if $DEVEL; #-+-> DEVEL ONLY <-+-#

        # Filter the message
        my $result = $self->{QUERY}->is_satisfied_by($filter->{'f'}, $alert_message);
        if ($result) {
            # Message matched filter in 'allow' chain, return authoritative 'allow'
            return (1, $filter->{'r'}, $filter->{'c'}) if $chain_type eq CHAIN_ALLOW;
            # Message matched filter in 'deny' chain, return authoritative 'deny'
            return (0, $filter->{'r'}, $filter->{'c'}) if $chain_type eq CHAIN_DENY;
        }
    }
    # Message did not matched any filter, we do not know at this point
    return (undef, undef, undef);
}

# _check_chain_anded($$$) [PROTECTED]
#
#  Usage     :
#  Purpose   : Filter the given message through selected filter chain
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _check_chain_anded($$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($filter_chain, $chain_type, $alert_message) = @_;

    # Walk through all filters in selected chain
    foreach my $filter (@$filter_chain) {
        print STDERR "Checking filter '$filter->{'r'}' in chain '$chain_type'\n" if $DEVEL; #-+-> DEVEL ONLY <-+-#

        # Filter the message
        my $result = $self->{QUERY}->is_satisfied_by($filter->{'f'}, $alert_message);
        unless ($result) {
            # Message did not matched some filter, we do not know at this point
            return (undef, undef, undef);
        }
    }
    # Message matched all filters in 'allow' chain, return authoritative 'allow'
    return (1, '__CHAIN__', 0) if $chain_type eq CHAIN_ALLOW;
    # Message matched all filters in 'deny' chain, return authoritative 'deny'
    return (0, '__CHAIN__', 0) if $chain_type eq CHAIN_DENY;
}

=pod

=back

=cut

1;
