package Data::Filter::Rule::OperationUnary;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Data::Filter::Rule::OperationUnary - Mentat message filter tree unary operation node

=head1 SYNOPSIS

 use Data::Filter::Rule::OperationUnary;

=head1 DESCRIPTION



=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1), Data::Filter::Rule(3), Data::Filter::Rule::Operation(3).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

#use Data::Dumper; #--- DEVELOPMENT ONLY ---#
#use Smart::Comments; #--- DEVELOPMENT ONLY ---#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Data::Filter::Rule::Operation;
use Data::Core qw(:c_dtypes :c_opers);

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant OP_NOT => 'NOT';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# List of valid operation types and their normalized names
my %OPERATIONS = (
        'not'   => OP_NOT,
        'NOT'   => OP_NOT,
    );

# List of appropriate result datatypes for given operations
my %DATATYPES = (
        OP_NOT, DT_BOOL,
    );

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL @ISA);
    $VERSION = '0.1';
    $DEVEL   = 0;
    @ISA     = ('Data::Filter::Rule::Operation');
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item valid_operation($) [PUBLIC]

  Usage     : my $operation = $rule->valid_operation();
  Purpose   :
  Returns   :
  Arguments :
  Throws    :
  Comments  :
  See Also  :

=cut

sub valid_operation($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($operation,) = @_;
    confess "Invalid unary operation '$operation'" unless $OPERATIONS{$operation};

    return ($OPERATIONS{$operation}, $DATATYPES{$OPERATIONS{$operation}});
}

#-------------------------------------------------------------------------------

=item left_operand($) [PUBLIC]

  Usage     :
  Purpose   :
  Returns   :
  Arguments :
  Throws    :
  Comments  :
  See Also  :

=cut

sub left_operand(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($operand,) = @_;

    # Setter mode
    if ($operand) {
        confess "Left operand must be Data::Filter::Rule" unless $operand->isa('Data::Filter::Rule');
        $self->{LEFT_OPERAND} = $operand;
    }
    return $self->{LEFT_OPERAND};
}

#-------------------------------------------------------------------------------

=item traverse($;@) [PUBLIC]

  Usage     : my $result = Data::Filter::Rule->traverse($traverser);
  Purpose   : Traversal through the filter rule tree with given traverser object
  Returns   :
  Arguments :
  Throws    :
  Comments  :
  See Also  :

=cut

sub traverse($;@)
{
    my $self    = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($traverser, @args) = @_;
    confess "Traverser object must be given as first argument" unless ref $traverser;

    my $lo = $self->left_operand();

    my $name = lcfirst($self->name());
    confess "Traverser object can`t handle '$name' callback hook" unless $traverser->can($name);

    my $lr = $lo->traverse($traverser, @args);

    return $traverser->$name($self, $self->operation(), $lo, $lr, @args);
}

#-------------------------------------------------------------------------------

=item to_string() [PUBLIC]

  Usage     : my $result = Data::Filter::Rule->to_string();
  Purpose   :
  Returns   :
  Arguments :
  Throws    :
  Comments  :
  See Also  :

=cut

sub to_string()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    return "(".$self->operation()." ".$self->left_operand().")";
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init($) [PROTECTED]
#
#  Usage     :
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class
#  Comments  :
#  See Also  :

sub _init($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($type, $left) = @_;
    confess "Unary operation must have one operand" unless ref $left;

    $self->operation($type);
    $self->left_operand($left);

    return $self;
}

=pod

=back

=cut

1;
