package Data::Path::Parser;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Data::Path::Parser - Provides methods for handling data paths (MPath)

=head1 SYNOPSIS

 use Data::Path::Parser;

 #--------------------------------------
 # Module usage via static class methods
 #--------------------------------------

 # Parse out the first node (chunk) from given data path:
 eval {
    my ($name, $type, $index, $rest) = Data::Path::Parser->parse_next_chunk($mpath);
 };
 if($@) { print "MESSAGE PATH ERROR: $@\n"; }

 # Parse the given data path chunk by chunk:
 my @result;
 eval {
    do {
        my ($name, $type, $index, $rest) = Data::Path::Parser->parse_next_chunk($mpath);
        push(@result, [$next_node_xpath, $name, $type, $index, $mpath]);
    } while ($mpath);
 };
 if($@) { print "MESSAGE PATH ERROR: $@\n"; }

 # Parse the given data path in one call:
 eval {
    my @chunks = Data::Path::Parser->parse($mpath);
 };
 if($@) { print "MESSAGE PATH ERROR: $@\n"; }

 # Parse the given data path in one call, but suppress all raised exceptions:
 my @chunks = Data::Path::Parser->parse_quiet($mpath);

 # Just validate the given data path
 my $result = Data::Path::Parser->validate($mpath);

 #----------------------------------
 # Module usage via instance methods
 #----------------------------------

 my $parser = new Data::Path::Parser();

 # Parse out the first node (chunk) from given data path:
 eval {
    my ($name, $type, $index, $rest) = $parser->parse_next_chunk($mpath);
 };
 if($@) { print "MESSAGE PATH ERROR: $@\n"; }

 # Parse the given data path chunk by chunk:
 my @result;
 eval {
    do {
        my ($name, $type, $index, $rest) = $parser->parse_next_chunk($mpath);
        push(@result, [$next_node_xpath, $name, $type, $index, $mpath]);
    } while ($mpath);
 };
 if($@) { print "MESSAGE PATH ERROR: $@\n"; }

 # Parse the given data path in one call:
 eval {
    my @chunks = $parser->parse($mpath);
 };
 if($@) { print "MESSAGE PATH ERROR: $@\n"; }

 # Parse the given data path in one call, but suppress all raised exceptions:
 my @chunks = $parser->parse_quiet($mpath);

 # Just validate the given data path
 my $result = $parser->validate($mpath);

=head1 DESCRIPTION

This module is used to parse data path, which can be then used
for manipulating the Mentat message objects. See the data path SYNTAX
section for more details on data path.

=head1 MESSAGE PATH SYNTAX (MPATH)

data path (MPath) syntax is a subset of XPath syntax. Is uses only basic
constructs for identifying the elements in the messages. Names of the nodes are defined
in the RFC 4765, however at this point of implementation this module does validity
check only on the first node in the path: it must be either 'Alert', or 'Heartbeat'.
Any other node (element or atribute) names or the path structure are not checked.

MPath is a string composed of node identifiers separated by '/' character.
Node identifier is composed of optional '@' character followed by node name
followed by optional integer index enclosed within '[' and ']' the braces.
Character '@' implies, that the node is attribute node and this node must be
the last node in the MPath (attribute nodes may not have any child nodes).
Optional integer index is valid only for element nodes. JSONPath standard allows
multiple occurences of certain element nodes, index allows to identify particular
one. DPath must not start with the '/' character.

Some valid MPath examples:
    Node/Name
    Source[2]/IP4[1]

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC, STATIC]

  Usage     : my $instance = new Data::Path::Parser();
  Purpose   : Create new instance of the MPath parser
  Returns   : Reference to the new instance of the MPath parser
  Arguments : None
  Throws    : Croaks, if invoked on object
  Comments  : All provided methods can be used either as class, or instance methods
  See Also  :

=cut

sub new
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my $self = bless ({}, $class);
    return $self;
}

=item parse_next_chunk($) [PUBLIC, HYBRID]

  Usage     : my ($name, $type, $index, $rest) = Data::Path::Parser->parse_next_chunk($mpath); or
              my ($name, $type, $index, $rest) = $parser->parse_next_chunk($mpath);
  Purpose   : Parse next chunk from the given data path
  Arguments : STRING $mpath - data path to parse
  Returns   : ARRAY (SUCCESS):
                STRING $name - name of the next node
                ENUM $type - type of the next node ('attr'|'elem')
                INTEGER $index - optional integer index of the node, or undef
                STRING $rest - rest of the data path to parse
  Throws    : Croaks, if not given any data path to parse
              Croaks, if node name syntax is invalid
              Croaks, if attribute node contains index (there may be only one instance of each attribute node)
              Croaks, if attribute node is not last node in path (attribute nodes may not have any child nodes)
  Comments  : This method is internally used by parse(), validate() and methods
  See Also  : parse(), validate(), methods

=cut

sub parse_next_chunk($)
{
    my $invocant = shift;
    my ($mpath,) = @_;
    confess "Invalid usage, data path must be given as argument" unless defined($mpath);

    # MPath separators are slashes '/', we are interested only in the next chunk
    my ($next_node, $rest) = split('/', $mpath, 2);
    croak ((caller(0))[3] . ": undefined next data path node") unless $next_node;

    # Node name must comply to this pattern:
    #  - first character may be '@', that means the name is the attribute node
    #  - node name itself may contain following characters: [-_:a-zA-Z]
    #  - after the node name, element index may be present: [1]
    if ($next_node =~ /^(@)?([-_:a-zA-Z0-9]+)(?:\[(\d+|\*|#)\])?$/) {
        my ($type, $name, $index) = ($1, $2, $3);
        $type = ($type) ? 'attr' : 'elem';

        # Index starts at 1, not 0
        croak ((caller(0))[3] . ": invalid index value '0', index starts at '1'") if defined $index and $index ne '*' and $index ne '#' and $index == 0;

        # Attribute node may not have indexes, there is always only one attribute instance
        croak ((caller(0))[3] . ": atribute node '$next_node' must have only one instance") if $type eq 'attr' and $index;

        # Attribute node must be the last one in a path, it may not contain any other nodes
        croak ((caller(0))[3] . ": atribute node '$next_node' may not contain any child nodes") if $type eq 'attr' and $rest;

        # We are okay, generate the next node chunk and return the results
        return ($name, $type, $index, $rest);
    }
    else {
        croak ((caller(0))[3] . ": invalid Mentat message node: '$next_node'");
    }
}

=item parse($) [PUBLIC, HYBRID]

  Usage     : my @result = Data::Path::Parser->parse($mpath); or
              my @result = $parser->parse($mpath);
  Purpose   : Parse the whole MPath in one call
  Returns   : array containing array references containing values returned by parse_next_chunk() method (SUCCESS)
  Arguments : Same as parse_next_chunk() method
  Throws    : Croaks, if first node of MPath is not 'Alert' or 'Heartbeat'
              Same as parse_next_chunk() method
  Comments  : Internally uses parse_next_chunk() method, see it for details
  See Also  : parse_next_chunk() method

=cut

sub parse($)
{
    my $invocant = shift;
    my ($mpath,) = @_;
    confess "Invalid usage, data path must be given as argument" unless defined($mpath);

    # Use the parse next chunk in a loop to parse the whole path
    my @result;
    my ($name, $type, $index);
    do {
        ($name, $type, $index, $mpath) = Data::Path::Parser->parse_next_chunk($mpath);
        push(@result, [$name, $type, $index, $mpath]);
    } while ($mpath);

    # Validate the first element of the data path - it must be either 'Alert', or 'Heartbeat'
    #croak ((caller(0))[3] . ": invalid Mentat message type '" . $result[0][0] . "', only 'Alert' and 'Heartbeat' are allowed\n") unless $result[0][0] =~ /^(Alert|Heartbeat)$/;

    return @result;
}

=item parse_quiet($) [PUBLIC, HYBRID]

  Usage     : my $result = Data::Path::Parser->parse_quiet($mpath); or
              my $result = $parser->parse_quiet($mpath);
  Purpose   : Same as parse() method, but does not throw any exceptions (except if not given any MPath to parse)
  Returns   : array containing array references containing values returned by parse_next_chunk() method (SUCCESS)
              undef (FAILURE)
  Arguments : Same as parse() method
  Throws    : Croaks, if not given any MPath to parse
  Comments  : Internally uses parse() method, see it for details
  See Also  : parse_next_chunk(), parse() methods

=cut

sub parse_quiet($)
{
    my $invocant = shift;
    my ($mpath,) = @_;
    confess "Invalid usage, data path must be given as argument" unless defined($mpath);

    # Try to parse the given data path
    my @result;
    eval {
        @result = Data::Path::Parser->parse($mpath);
    };
    # Any exceptions caught mean that there was some error in the data path syntax
    if ($@) { return undef; }
    else    { return @result; }
}

=item join($$$) [PUBLIC, HYBRID]

  Usage     : my $mpath = Data::Path::Parser->join($mpath, $name, $index); or
              my $mpath = $parser->join($mpath, $name, $index);
  Purpose   : Join the chunk of data path with given MPath
  Arguments : STRING $mpath - data path to parse [MANDATORY]
              STRING $name - name of the next node [MANDATORY]
              INTEGER $index - optional integer index of the node, or undef
  Returns   : STRING $mpath
  Throws    : Croaks, if not given any data path to parse
              Croaks, if node name syntax is invalid
              Croaks, if attribute node contains index (there may be only one instance of each attribute node)
              Croaks, if attribute node is not last node in path (attribute nodes may not have any child nodes)
  Comments  : This method is internally used by parse(), validate() and methods
  See Also  : parse(), validate(), methods

=cut

sub join($$$)
{
    my $invocant = shift;
    my ($mpath, $name, $index) = @_;
    confess "Invalid usage, data path must be given as argument" unless defined($mpath);
    confess "Invalid usage, path chunk name must be given as argument" unless defined($name);

    return ($mpath?"$mpath/":'') . $name . ($index?"[$index]":'');
}

=item validate($) [PUBLIC, HYBRID]

  Usage     : my $result = Data::Path::Parser->validate($mpath); or
              my $result = $parser->validate($mpath);
  Purpose   : Validate, if the given MPath is syntactically correct
  Returns   : integer 1 (SUCCESS)
              integer 0 (FAILURE)
  Arguments : Same as parse() method
  Throws    : Nothing
  Comments  : Internally uses parse() method, see it for details
  See Also  : parse() method

=cut

sub validate($)
{
    my $invocant = shift;
    my ($mpath,) = @_;
    confess "Invalid usage, data path must be given as argument" unless defined($mpath);

    # Try to parse the given data path
    eval {
        Data::Path::Parser->parse($mpath);
    };
    # Any exceptions caught mean that there was some error in the MPath syntax
    if ($@) { return 0; }
    else    { return 1; }
}

=pod

=back

=cut

1;
