package Mentat::Message::Uniter;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Message::Uniter - Short module description

=head1 SYNOPSIS

 use Mentat::Message::Uniter;

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

Jan Mach
Cesnet, z.s.p.o
jan.mach@cesnet.cz
http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Config::Parser;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant OP_SINGLE => 'SINGLE';
use constant OP_MULTI  => 'MULTI';

use constant OP_NODE_ADD => 'nadd';

use constant MODE_ALTER => 'ALTER';
use constant MODE_BUILD => 'BUILD';
use constant MODE_CLONE => 'CLONE';

use constant DFLT_MODE => MODE_ALTER();

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# Meta structure describing available aggregations
my %OPERATIONS = (
        'sum'   => [OP_SINGLE(), sub { my ($l, $r) = @_; return ($l + $r); }],
        'sub'   => [OP_SINGLE(), sub { my ($l, $r) = @_; return ($l - $r); }],
        'mul'   => [OP_SINGLE(), sub { my ($l, $r) = @_; return ($l * $r); }],
        'div'   => [OP_SINGLE(), sub { my ($l, $r) = @_; return ($l / $r); }],
        'mod'   => [OP_SINGLE(), sub { my ($l, $r) = @_; return ($l % $r); }],
        'con'   => [OP_SINGLE(), sub { my ($l, $r) = @_; return ("$l, $r"); }],
        'min'   => [OP_SINGLE(), sub { my ($l, $r) = @_; return $l unless defined $r; return $r unless defined $l; return ($l <= $r)?$l:$r; }],
        'max'   => [OP_SINGLE(), sub { my ($l, $r) = @_; return $l unless defined $r; return $r unless defined $l; return ($l >= $r)?$l:$r; }],
        'smin'  => [OP_SINGLE(), sub { my ($l, $r) = @_; return $l unless defined $r; return $r unless defined $l; return (lc($l) le lc($r))?$l:$r; }],
        'smax'  => [OP_SINGLE(), sub { my ($l, $r) = @_; return $l unless defined $r; return $r unless defined $l; return (lc($l) ge lc($r))?$l:$r; }],
        'inc'   => [OP_SINGLE(), sub { my ($l, $r) = @_; return ($l + 1); }],
        'frst'  => [OP_SINGLE(), sub { my ($l, $r) = @_; return $r unless defined $l; return $l; }],
        'last'  => [OP_SINGLE(), sub { my ($l, $r) = @_; return $l unless defined $r; return $r; }],
        'join'  => [OP_MULTI(),  sub { my ($l, $r) = @_; return [sort (@$l, @$r)]; }],
        'ujoin' => [OP_MULTI(),  sub { my ($l, $r) = @_; my %t = map { $_ => 1 } (@$l, @$r); return [sort keys(%t)]; }],
    );

my %DFLTS = (
    'sum'   => 0,
    'sub'   => 0,
    'mul'   => 0,
    'div'   => 0,
    'mod'   => 0,
    'con'   => '',
    'min'   => undef,
    'max'   => undef,
    'smin'  => undef,
    'smax'  => undef,
    'inc'   => 1,
    'frst'  => undef,
    'last'  => undef,
    'join'  => [],
    'ujoin' => [],
);

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC,STATIC]

  Usage     : my $instance = Mentat::Message::Uniter->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Mentat::Message::Uniter reference on success, croaks on failure
  Throws    : Croaks, if invoked on object
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my $self = bless ({}, $class);
    return $self->_init(@_);
}

=item unite($$;$) [PUBLIC]

  Usage     : my $result = $instance->unite($ruleset, $messages, $mode);
  Purpose   : Aggregate list of messages into the one according to the given ruleset
  Arguments : ARRAY REFERENCE $ruleset - aggregation ruleset
              ARRAY of Mentat::Messages
  Returns   : Mentat::Message reference on success, croaks on failure
  Throws    : Croaks, if invoked on object
              Croaks on aggregation failure

=cut

sub unite($$;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($ruleset, $messages, $mode) = @_;
    confess "Operation ruleset must be given as array reference" unless $ruleset and ref $ruleset eq 'ARRAY';
    confess "United messages must be passed as array reference" unless $messages and ref $messages eq 'ARRAY';
    confess "You have to suply at least two messages for unition" unless scalar(@$messages) >= 2;
    $mode = ($mode)?$mode:DFLT_MODE();

    my $target = shift @$messages;
    my $result = $target;
    if ($mode eq MODE_CLONE()) {
        $result = clone($result); $mode = MODE_ALTER();
    }
    foreach my $message (@$messages) {
        $result = $self->unite_into($ruleset, $result, $message);
    }
    return $result;
}

=item unite_into($$$) [PUBLIC]

  Usage     : my $result = $instance->unite_into($ruleset, $target, $source);
  Purpose   : Aggregate one message into another
  Arguments :
  Returns   : Mentat::Message reference on success, croaks on failure
  Throws    : Dies, if invoked on object

=cut

sub unite_into($$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($ruleset, $target, $source) = @_;
    confess "Operation ruleset must be given as array reference" unless $ruleset and ref $ruleset eq 'ARRAY';
    confess "You have to suply two messages for aggregation" unless $target and $source;

    my ($type, $left, $right, $op, $src_path, $dst_path, $value);
    foreach my $rule (@$ruleset) {
        ($op, $src_path, $dst_path) = @$rule;
        unless ($op eq OP_NODE_ADD()) {
            $dst_path = ($dst_path)?$dst_path:$src_path;
            $type = $OPERATIONS{$op}->[0];
            if ($type eq OP_SINGLE()) {
                $left = $target->path_value($src_path);
                $right = $source->path_value($src_path);
            }
            else {
                $left = $target->path_values($src_path);
                $right = $source->path_values($src_path);
            }
            $value = $self->_unite($op, $left, $right);
            ### Setting value: $value
            ### Setting value to path: $dst_path
            $target->path_node_set($dst_path, $value);
        }
        else {
            $target->path_node_set($src_path, $dst_path);
        }
    }
    return $target;
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Message::Uniter instance
#  Arguments :
#  Returns   : Mentat::Message::Uniter reference
#  Throws    : Croaks, if invoked on class
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.
#  See Also  :

sub _init
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self;
}

# _unite($) [PROTECTED]
#
#  Usage     : my $unite = $instance->_unite($ag, $left, $right);
#  Purpose   :
#  Arguments :
#  Returns   :
#  Throws    : Croaks, if invoked on class

sub _unite($$$)
{
    my $invocant = shift;
    my ($op, $left, $right) = @_;
    confess "Merge operation must be given as argument" unless defined($op);
    confess "Invalid operation '$op'" unless exists $OPERATIONS{$op};

    $left  = $DFLTS{$op} unless defined $left;
    $right = $DFLTS{$op} unless defined $right;

    return &{$OPERATIONS{$op}->[1]}($left, $right);
}

=pod

=back

=cut

1;
