package Mentat::Processor;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Processor - FRONT CONTROLLER - Main code logic for all MENTAT system components

=head1 SYNOPSIS

=head1 DESCRIPTION

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';
use Time::HiRes;

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Log::Loger;
use Value::Convertor;
use App::Feature::Daemon;
use App::Feature::Loger;
use Mentat::Core qw(:c_general :c_processor);
use Mentat::Processor::Component;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant D_E => 'emitters';
use constant D_H => 'handlers';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

# List of valid additional signals and their appropriate emittd events
my %VALID_SIGNALS = (
        'USR1'  =>  'signal_usr1',
        'USR2'  =>  'signal_usr2',
    );

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 1;
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

# Operator overloading
use overload '""' => 'to_string';

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new() [PUBLIC, STATIC]

  Usage     : my $processor = Mentat::Processor->new();
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Mentat::Processor reference on success, confesses on failure
  Throws    : Dies, if not invoked on class
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new()
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    ### Creating Mentat.Processor instance...
    my $self = bless ({}, $class);
    return $self->_init(@_);
}

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if invoked on class

=cut

sub DESTROY()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    ### Destroying Mentat.Processor instance...
    # This is vital to break the reference cycles, don`t leave any chances
    delete $self->{SCHEDULER};
    delete $self->{SCHEDULER_PRI};
    delete $self->{COMPONENTS};
    #delete $self->{COMPONENT_MAP}; # Removed in _cleanup() bootsrap sequence
    delete $self->{HANDLERS};
    #delete $self->{LINKS};         # Removed in _cleanup() bootsrap sequence
    delete $self->{STATISTICS};
}

=item to_string() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub to_string()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    #return blessed($self);
    return 'processor:controller';
}

=item start_time(;$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub start_time(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    confess "Method called too soon, not in run cycle yet" unless $self->{TIMER_RUN};
    return $self->{TIMER_RUN};
}

=item duration(;$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub duration(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    confess "Method called too soon, not in run cycle yet" unless $self->{TIMER_RUN};
    my ($as_string,) = @_;
    my $duration = time - $self->{TIMER_RUN};

    return $duration unless $as_string;
    return Value::Convertor->duration_to_string($duration);
}

#-------------------------------------------------------------------------------

=item bootstrap($) [PUBLIC]

  Usage     : $processor->bootstrap($CONFIG);
  Purpose   : Bootstrap all processor components
  Arguments : Hash containing application configurations
  Returns   : Nothing on success, dies on failure
  Throws    : Dies, if not invoked on instance
  Comments  : Internally uses private _configure_*() and _check() methods
  See Also  : Private _configure_*() and _check() methods

=cut

sub bootstrap($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($config,) = @_;
    confess "Bootstrap configurations must be passed as argument" unless $config;
    confess "Bootstrap configurations must be passed via HASH reference" unless ref $config eq 'HASH';

    # Many arrors may occur during the bootstrap process, catch them
    eval {
        ### Configuring message loger
        $self->_configure_loger($config->{CSECTION_LOGER()});
        $self->_sequence_begin('bootstrap', 'Initiating bootstrap sequence');

        ### Configuring main processor
        $self->_configure_processor($config->{CSECTION_PROCESSOR()});

        ### Configuring event scheduler
        $self->_configure_schedulers($config->{CSECTION_SCHEDULER()});

        ### Configuring processor components
        $self->_configure_components($config->{CSECTION_COMPONENTS()});

        ### Configuring inter-component links
        $self->_configure_links();

        ### Checking processor and processor components for sanity
        $self->_check();

        ### Daemonizing application
        $self->_daemonize($config->{CSECTION_DAEMON()});

        ### Cleaning up
        $self->_cleanup();

        # Set the initialized flag to true
        $self->{BOOTSTRAPPED} = 1;

        $self->_sequence_end('bootstrap', 'Bootstrap sequence completed');
        $self->_dump_setup();
    };
    # Trap and handle all errors during initializations
    if ($@) {
        print STDERR "[PCS] Bootstrap error: $@\n" unless Log::Loger->channel_count();
        Log::Loger->log_exit(__PACKAGE__, 'EMERG', "[PCS] Bootstrap error: $@");
    }
}

=item run(;$) [PUBLIC]

  Usage     : $processor->run($config);
  Purpose   : Run the application
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if invoked on class
  Comments  : This method MUST NOT return until the application is done
  See Also  :

=cut

sub run(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($config,) = @_;
    confess "Runtime configurations must be passed via HASH reference" if $config and ref $config ne 'HASH';

    # Provide default runtime configuration values
    $config = {} unless $config;

    # Start the RUN timer
    $self->{TIMER_RUN} = time;

    # Initiate the startup process, trap and handle all errors
    eval {
        $self->_sequence_begin('startup', 'Initiating startup sequence');
        $self->_startup($config->{CSECTION_STARTUP()});
        $self->_sequence_end('startup', 'Startup sequence complete');
    };
    if ($@) {
        print STDERR "[PCS] Startup error: $@\n" unless Log::Loger->channel_count();
        Log::Loger->log_exit(__PACKAGE__, 'EMERG', "[PCS] Startup error: $@");
    }

    # Initiate the event processing loop, trap and handle all errors
    eval {
        $self->_sequence_begin('eloop', 'Initiating event processing loop');
        $self->_loop($config->{CSECTION_LOOP()});
        $self->_sequence_end('eloop', 'Event processing loop complete');
    };
    if ($@) {
        print STDERR "[PCS] Runtime error: $@\n" unless Log::Loger->channel_count();
        Log::Loger->log_exit(__PACKAGE__, 'EMERG', "[PCS] Runtime error: $@");
    }

    # Initiate the shutdown process, trap and handle all errors
    eval {
        $self->_sequence_begin('shutdown', 'Initiating shutdown sequence');
        $self->_shutdown($config->{CSECTION_SHUTDOWN()});
        $self->_sequence_end('shutdown', 'Shutdown sequence complete');
    };
    if ($@) {
        print STDERR "[PCS] Shutdown error: $@\n" unless Log::Loger->channel_count();
        Log::Loger->log_exit(__PACKAGE__, 'EMERG', "[PCS] Shutdown error: $@");
    }

    # Try to make sure, that admin knows about shutdown
    my $duration = '... ('.Value::Convertor->duration_to_string($self->statistics->{'ts_eloop_d'}).')';
    my $message = ''; $message = join("\n", @{$self->{STATUS_MSGS}}) if @{$self->{STATUS_MSGS}};

    # Store the runlog
    $self->runlog_store();

    Log::Loger->log(__PACKAGE__, 'EMAIL', "[PCS] ********** Processing done, terminating $duration **********\n$message");
}

#-------------------------------------------------------------------------------

=item register_event($$) [PUBLIC]

  Usage     : $processor->register_event($emitter_obj, $event_name);
  Purpose   : Register new event to the application
  Arguments : BLESSED $emitter_obj - object emitting the event
              STRING $event_name - name of the emitted event
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance or given invalid arguments

=cut

sub register_event($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    confess "Can't register new event after completing application bootstrap sequence" if $self->{BOOTSTRAPPED};
    my ($emitter_obj, $event_name) = @_;
    confess "Event emitting object must be passed as first argument" unless $emitter_obj and blessed($emitter_obj);
    confess "Event name must be passed as second argument" unless $event_name;

    Log::Loger->debug(__PACKAGE__, "[PCS] Registering event '$event_name' emited by '$emitter_obj'");
    push(@{$self->{HANDLERS}->{$event_name}->{D_E()}}, $emitter_obj);
}

=item register_event_handler($$;$) [PUBLIC]

  Usage     : $processor->register_event($handler_obj, $event_name, $handler_name);
  Purpose   : Register new event handler to the application
  Arguments : BLESSED $handler_obj - object handling the event
              STRING $event_name - name of the handled event
              STRING $handler_name - name of the event handler
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance or given invalid arguments

=cut

sub register_event_handler($$;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    confess "Can't register new event handler after completing application bootstrap sequence" if $self->{BOOTSTRAPPED};
    my ($handler_obj, $event_name, $handler_name) = @_;
    confess "Event handling object must be passed as first argument" unless $handler_obj and blessed($handler_obj);
    confess "Event name must be passed as second argument" unless $event_name;

    # By default handler name is the same as event name
    $handler_name = $handler_name || $event_name;
    confess "Handling object '$handler_obj' is not capable of handling '$event_name', event handler '$handler_name' is missing" unless $handler_obj->can($handler_name);

    Log::Loger->debug(__PACKAGE__, "[PCS] Registering handler '$handler_name' of object '$handler_obj' for event '$event_name'");
    push(@{$self->{HANDLERS}->{$event_name}->{D_H()}}, [$handler_obj, $handler_name]);
}

=item register_linked_component($$;$) [PUBLIC]

  Usage     : $processor->register_linked_component($component, $linked, $id);
  Purpose   : Register new link between application components
  Arguments : Mentat::Processor::Component $component - component requesting the link
              STRING $linked - specification of the linked component
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance or given invalid arguments

=cut

sub register_linked_component($$;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    confess "Can't register new component link after completing application bootstrap sequence" if $self->{BOOTSTRAPPED};
    my ($component, $linked, $id) = @_;
    confess "Component requesting the link must be passed as first argument" unless $component and blessed($component);
    confess "Linked component specification must be passed as second argument" unless defined $linked;

    Log::Loger->debug(__PACKAGE__, "[PCS] Registering link '$linked' for component '$component'");
    $self->{LINKS}->{$component->id()}->{$linked} = ($id)?$id:1;
}

#-------------------------------------------------------------------------------

=item emit_event($;@) [PUBLIC]

  Usage     : my $id = $processor->emit_event($event, @args);
  Purpose   : Emit new event and schedule it at the end of event queue
  Arguments : STRING $event - event name [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : MIXED $id unique event identifier
  Throws    : Dies, if not invoked on class or given invalid arguments

=cut

sub emit_event($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($event_name, @args) = @_;
    confess "Event name must be passed as first argument" unless $event_name;
    confess "Attempted to emit unknown event '$event_name'" unless exists $self->{HANDLERS}->{$event_name};

    ### emit_event: $event_name
    $self->statistics->{"ee_$event_name"}++;
    Log::Loger->debug(__PACKAGE__, "[PCS] Emitting event '$event_name', scheduled at the queue end");
    return $self->{SCHEDULER}->schedule($event_name, @args);
}

=item emit_event_f($;@) [PUBLIC]

  Usage     : my $id = $processor->emit_event_f($event, @args);
  Purpose   : Emit new event and schedule it at the beginning of event queue
  Arguments : STRING $event - event name [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : MIXED $id unique event identifier
  Throws    : Dies, if not invoked on class or given invalid arguments

=cut

sub emit_event_f($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($event_name, @args) = @_;
    confess "Event name must be passed as first argument" unless $event_name;
    confess "Attempted to emit unknown event '$event_name'" unless exists $self->{HANDLERS}->{$event_name};

    ### emit_event_f: $event_name
    $self->statistics->{"ef_$event_name"}++;
    Log::Loger->debug(__PACKAGE__, "[PCS] Emitting event '$event_name', scheduled at the queue beginning");
    return $self->{SCHEDULER}->schedule_first($event_name, @args);
}

=item emit_event_at($$;@) [PUBLIC]

  Usage     : my $id = $processor->emit_event_at($event, @args);
  Purpose   : Emit new event and schedule it at specific time
  Arguments : STRING $event - event name [MANDATORY]
              INTEGER $timestamp - unix timestamp [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : MIXED $id unique event identifier
  Throws    : Dies, if not invoked on class or given invalid arguments

=cut

sub emit_event_at($$;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($event_name, $timestamp, @args) = @_;
    confess "Event name must be passed as first argument" unless $event_name;
    confess "Event timestamp must be passed as second argument" unless $timestamp and int($timestamp);
    confess "Attempted to emit unknown event '$event_name'" unless exists $self->{HANDLERS}->{$event_name};

    ### emit_event_at: $event_name
    $self->statistics->{"ea_$event_name"}++;
    Log::Loger->debug(__PACKAGE__, "[PCS] Emitting event '$event_name', scheduled at '$timestamp'");
    return $self->{SCHEDULER}->schedule_at($event_name, $timestamp, @args);
}

=item emit_event_pri($;@) [PUBLIC]

  Usage     : my $id = $processor->emit_event_pri($event, @args);
  Purpose   : Emit new priority event and schedule it at the end of event queue
  Arguments : STRING $event - event name [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : MIXED $id unique event identifier
  Throws    : Dies, if not invoked on class or given invalid arguments

=cut

sub emit_event_pri($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($event_name, @args) = @_;
    confess "Event name must be passed as first argument" unless $event_name;
    confess "Attempted to emit unknown event '$event_name'" unless exists $self->{HANDLERS}->{$event_name};

    ### emit_event: $event_name
    $self->statistics->{"eep_$event_name"}++;
    Log::Loger->debug(__PACKAGE__, "[PCS] Emitting priority event '$event_name', scheduled at the queue end");
    return $self->{SCHEDULER_PRI}->schedule($event_name, @args);
}

=item emit_event_pri_f($;@) [PUBLIC]

  Usage     : my $id = $processor->emit_event_pri_f($event, @args);
  Purpose   : Emit new priority event and schedule it at the beginning of event queue
  Arguments : STRING $event - event name [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : MIXED $id unique event identifier
  Throws    : Dies, if not invoked on class or given invalid arguments

=cut

sub emit_event_pri_f($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($event_name, @args) = @_;
    confess "Event name must be passed as first argument" unless $event_name;
    confess "Attempted to emit unknown event '$event_name'" unless exists $self->{HANDLERS}->{$event_name};

    ### emit_event_f: $event_name
    $self->statistics->{"efp_$event_name"}++;
    Log::Loger->debug(__PACKAGE__, "[PCS] Emitting priority event '$event_name', scheduled at the queue beginning");
    return $self->{SCHEDULER_PRI}->schedule_first($event_name, @args);
}

=item emit_event_pri_at($$;@) [PUBLIC]

  Usage     : my $id = $processoremit_event_pri_at($event, @args);
  Purpose   : Emit new priority event and schedule it at specific time
  Arguments : STRING $event - event name [MANDATORY]
              INTEGER $timestamp - unix timestamp [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : MIXED $id unique event identifier
  Throws    : Dies, if not invoked on class or given invalid arguments

=cut

sub emit_event_pri_at($$;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($event_name, $timestamp, @args) = @_;
    confess "Event name must be passed as first argument" unless $event_name;
    confess "Event timestamp must be passed as second argument" unless $timestamp and int($timestamp);
    confess "Attempted to emit unknown event '$event_name'" unless exists $self->{HANDLERS}->{$event_name};

    ### emit_event_at: $event_name
    $self->statistics->{"eap_$event_name"}++;
    Log::Loger->debug(__PACKAGE__, "[PCS] Emitting priority event '$event_name', scheduled at '$timestamp'");
    return $self->{SCHEDULER_PRI}->schedule_at($event_name, $timestamp, @args);
}

=item execute_event($;@) [PUBLIC]

  Usage     : my $id = $processorexecute_event($event, @args);
  Purpose   : Bypass event emission and immediatelly start to execute chain of the
              given event
  Arguments : STRING $event - event name [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : BOOL $result - TRUE on SUCCESS, FALSE on FAILURE
  Throws    : Dies, if not invoked on class or given invalid arguments

=cut

sub execute_event($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($event_name, @args) = @_;
    confess "Event name must be passed as first argument" unless $event_name;
    confess "Attempted to execute unknown event '$event_name'" unless exists $self->{HANDLERS}->{$event_name};

    ### Executing event: $event_name
    $self->statistics->{"ex_$event_name"}++;
    my ($ehobj, $ehname, $cont);
    my $ehchain = $self->{HANDLERS}->{$event_name}->{D_H()};
    foreach my $eh (@$ehchain) {
        ($ehobj, $ehname) = @$eh;
        Log::Loger->debug(__PACKAGE__, "[PCS] Executing handler '$ehname' for event '$event_name' on component '$ehobj'");
        ($cont, @args) = $ehobj->$ehname(@args);
        last unless $cont eq EC_CONTINUE;
    }
    return ($cont, @args);
}

#-------------------------------------------------------------------------------

=item statistics() [PUBLIC]

  Usage     : my $statistics = $processor->statistics();
  Purpose   : Get processing statistics
  Arguments : None
  Returns   : HASH REFERENCE $statistics with various keys
  Throws    : Dies, if not invoked on class
  Comments  : Mainly for debugging, development and optimization purposes

=cut

sub statistics()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->{STATISTICS};
}

=item runlog_set($) [PUBLIC]

  Usage     : my $runlog = $processor->runlog_set($)
  Purpose   : Set processing runlog
  Arguments : HASH REFERENCE $runlog
  Returns   : HASH REFERENCE $runlog
  Throws    : Dies, if not invoked on class
  Comments  : Mainly for debugging, development and optimization purposes

=cut

sub runlog_set($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($rl,) = @_;
    return $self->{RUNLOG} = $rl;
}

=item runlog_get() [PUBLIC]

  Usage     : my $runlog = $processor->runlog_get($)
  Purpose   : Get processing runlog
  Arguments : None
  Returns   : HASH REFERENCE $runlog
  Throws    : Dies, if not invoked on class
  Comments  : Mainly for debugging, development and optimization purposes

=cut

sub runlog_get()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    # Lazy init
    $self->{RUNLOG} = {} unless $self->{RUNLOG};

    return $self->{RUNLOG};
}

=item runlog_store() [PUBLIC]

  Usage     : my $runlog = $processor->runlog_store()
  Purpose   : Store processing runlog
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on class
  Comments  : Mainly for debugging, development and optimization purposes

=cut

sub runlog_store()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    if ($self->{RUNLOG}) {
        $self->{RUNLOG}->{app_name} = Mentat::Core::app_name() unless exists $self->{RUNLOG}->{app_name};
        $self->{RUNLOG}->{ts_start} = $self->start_time()      unless exists $self->{RUNLOG}->{ts_start};
        $self->{RUNLOG}->{ts}       = time()                   unless exists $self->{RUNLOG}->{ts};
        Mentat::Core::json_save($self->{RUNLOG}, $self->{RUNLOG_FILE});
    }
}

=item done(;$$) [PUBLIC]

  Usage     : $processor->done(1);               # Set the flag
              until ($processor->done()) { ... } # Work until done
  Purpose   : After setting up of this flag, processor MUST gracefully exit
              the run() method as soon as possible.
  Arguments : BOOL $done_flag - any perlish TRUE value will turn the flag ON
  Returns   : INTEGER 1, if the processor should stop, INTEGER 0 otherwise
  Throws    : Dies, if not invoked on instance
  Comments  : Once the done flag is set, it can`t be cleared
  See Also  : run() method

=cut

sub done(;$$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($done_flag, $origin) = @_;

    if ($done_flag) {
        $self->{DONE}++;
        $origin = ($origin)?($origin):((caller(1))[3]);
        #Log::Loger->info(__PACKAGE__, "[PCS] DONE flag set by '$origin', stack backtrace" . Carp::longmess());
        Log::Loger->info(__PACKAGE__, "[PCS] DONE flag set by '$origin'");
    }
    return $self->{DONE};
}

#-------------------------------------------------------------------------------

sub broadcast($@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($method, @args) = @_;

    foreach my $component ($self->_component_list()) {
        $component->$method(@args);
    }
}

#-------------------------------------------------------------------------------

sub pause()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->broadcast('state', STATE_PAUSED);
}
sub resume()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->broadcast('state', STATE_NORMAL);
}

#-------------------------------------------------------------------------------

=item wait($) [PUBLIC]

  Usage     : $processor->wait($time);
  Purpose   : Wait for given amount of time (sleep)
  Arguments : FLOAT $time - how long to wait
  Returns   : FLOAT $slept how long actually waited
  Throws    : Dies, if not invoked on instance

=cut

sub wait($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($time,) = @_;
    confess "Waiting time must be greater than 0 seconds" unless $time and $time > 0;

    return Time::HiRes::sleep($time);
}

#-------------------------------------------------------------------------------

=item status_msg($) [PUBLIC]

  Usage     : $processor->status_msg($message);
  Purpose   : Setup the status message, which will be sent to the user after processing is done
  Arguments : STRING $message - status message [MANDATORY]
  Returns   : Nohing
  Throws    : Dies, if not invoked on instance

=cut

sub status_msg($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($message,) = @_;
    confess "Status message must be given as argument" unless $message;

    push(@{$self->{STATUS_MSGS}}, $message);
}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init() [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Processor instance
#  Arguments : None
#  Returns   : Mentat::Processor reference $self
#  Throws    : Dies, if not invoked on instance

sub _init()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    # Event scheduler
    $self->{SCHEDULER} = undef;
    # List of procesor components
    $self->{COMPONENTS} = {};
    # Component map according to the types, names and type:names
    $self->{COMPONENT_MAP} = {};
    # Event handler map
    $self->{HANDLERS} = {};
    # Component link map
    $self->{LINKS} = {};
    # Processor statistics
    $self->{STATISTICS} = {};
    # Status messages
    $self->{STATUS_MSGS} = [];

    return $self;
}

# _generate_component_id() [PROTECTED]
#
#  Usage     : my $id = $processor->_generate_component_id();
#  Purpose   : Generate unique component identifier
#  Arguments : None
#  Returns   : INTEGER $id
#  Throws    : Dies, if not invoked on instance

sub _generate_component_id()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return ++$self->{ID_GEN};
}

# _component_list(;$) [PROTECTED]
#
#  Usage     : my @list = $processor_component_list(;$);
#  Purpose   : Return list of all components in order they were configured
#  Arguments : BOOL $reverse - reverse the order [OPTIONAL]
#  Returns   : ARRAY @components containing Mentat::Processor::Component references
#  Throws    : Dies, if not invoked on instance

sub _component_list(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($reverse,) = @_;

    unless ($reverse) {
        return sort {$a->id()<=>$b->id()} values(%{$self->{COMPONENTS}});
    }
    else {
        return sort {$b->id()<=>$a->id()} values(%{$self->{COMPONENTS}});
    }
}

#-------------------------------------------------------------------------------

# _configure_loger(;$) [PROTECTED]
#
#  Usage     : $self->_configure_loger($config);
#  Purpose   : Configure message loger
#  Arguments : HASH REFERENCE $config - loger configurations
#  Returns   : Mentat::Processor reference $self
#  Throws    : Dies, if not invoked on instance
#

sub _configure_loger(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($config,) = @_;

    # Logging is optional
    if ($config) {
        die "Loger configurations must be passed via ARRAY reference in '".CSECTION_LOGER."' configuration section\n" unless ref $config eq 'HASH';
        die "Log channels must be specified in '".CSECTION_LOGER."' configuration section under key '".CKEY_LOG."'\n" unless exists $config->{CKEY_LOG()};
        die "Log channels must be passed as array reference\n" unless ref $config->{CKEY_LOG()} eq 'ARRAY';

        App::Feature::Loger->initiate(@{$config->{CKEY_LOG()}});

        Log::Loger->info(__PACKAGE__, "[PCS] Successfully initialized message loging services");
        Log::Loger->info(__PACKAGE__, "[PCS] Using file '".Mentat::Core::detect_config_file()."' as main application configuration file");
    }

    return $self;
}

# _configure_processor(;$) [PROTECTED]
#
#  Usage     : $self->_configure_processor($config);
#  Purpose   : Configure main processor
#  Arguments : HASH REFERENCE $config - processor configurations
#  Returns   : Mentat::Processor reference $self
#  Throws    : Dies, if not invoked on instance
#

sub _configure_processor(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($config,) = @_;

    # No processor configuration is mandatory
    if ($config) {
        die "Processor configurations must be passed via HASH reference in '".CSECTION_PROCESSOR."' configuration section\n" unless ref $config eq 'HASH';
        #Log::Loger->info(__PACKAGE__, "[PCS] Processor configurations: ".Dumper($config));

        # Setup name of the runlog file
        $self->{RUNLOG_FILE} = $config->{CKEY_RUNLOG()} if ($config->{CKEY_RUNLOG()});

        # Change the application path
        if ($config->{CKEY_PATH()}) { $config->{CKEY_PATH()} =~ /(.*)/; $ENV{PATH} = $1; }

        # Install default signal handlers
        $SIG{INT}  = $SIG{QUIT} = $SIG{TERM} = sub { $self->done(1, 'SIGNAL'); };
        $SIG{KILL} = sub { die "Brutal KILL signal received !!!\n"; };
        $SIG{PIPE} = 'IGNORE';

        # Install additional signal handlers
        if ($config->{CKEY_SIGNALS()}) {
            foreach my $signal (@{$config->{CKEY_SIGNALS()}}) {
                die "Invalid additional signal '$signal', posible values are '".join(',',keys(%VALID_SIGNALS))."'\n" unless $VALID_SIGNALS{$signal};
                $SIG{$signal} = sub { $self->emit_event_pri($VALID_SIGNALS{$signal}); };
                $self->register_event($self, $VALID_SIGNALS{$signal});
            }
        }

        Log::Loger->info(__PACKAGE__, "[PCS] Successfully configured main processor");
    }
    return $self;
}

# _configure_shedulers($) [PROTECTED]
#
#  Usage     : $self->_configure_shedulers($config);
#  Purpose   : Configure event schedulers
#  Arguments : HASH REFERENCE $config - scheduler configurations
#  Returns   : Mentat::Processor reference $self
#  Throws    : Dies, if not invoked on instance
#              Dies, if given invalid arguments

sub _configure_schedulers($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($config,) = @_;
    die "Scheduler configurations must be passed via HASH reference in '".CSECTION_SCHEDULER."' configuration section\n" unless ref $config eq 'HASH';
    die "Scheduler class name must be specified in '".CSECTION_SCHEDULER."' configuration section under key '".CKEY_CLASS."'\n" unless exists $config->{CKEY_CLASS()};

    my $s_class = $config->{CKEY_CLASS()};
    my $s_obj = $s_class->new(%$config);
    die "Scheduler must be instance of 'Mentat::Processor::Scheduler'\n" unless $s_obj->isa('Mentat::Processor::Scheduler');

    my $s_obj_pri = $s_class->new(%$config);
    die "Priority scheduler must be instance of 'Mentat::Processor::Scheduler'\n" unless $s_obj_pri->isa('Mentat::Processor::Scheduler');

    $self->{SCHEDULER}     = $s_obj;
    $self->{SCHEDULER_PRI} = $s_obj_pri;

    Log::Loger->info(__PACKAGE__, "[PCS] Successfully configured event schedulers");

    return $self;
}

# _configure_components($) [PROTECTED]
#
#  Usage     : $self->_configure_components($config);
#  Purpose   : Configure processor components
#  Arguments : HASH REFERENCE $config - component configurations
#  Returns   : Mentat::Processor reference $self
#  Throws    : Dies, if not invoked on instance
#              Dies, if given invalid arguments

sub _configure_components($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($config,) = @_;
    die "Component configurations must be passed via ARRAY reference in '".CSECTION_COMPONENTS."' configuration section\n" unless ref $config eq 'ARRAY';

    my ($c_class, $c_id, $c_obj, $c_type, $c_name);
    foreach my $c_cfg (@$config) {
        die "Component configuration must be passed via HASH reference in list in '".CSECTION_COMPONENTS."' configuration section\n" unless ref $c_cfg eq 'HASH';
        die "Component class name must be specified in '".CSECTION_COMPONENTS."' configuration section under key '".CKEY_CLASS."'\n" unless exists $c_cfg->{CKEY_CLASS()};

        $c_class = $c_cfg->{CKEY_CLASS()};
        $c_id    = $self->_generate_component_id();
        $c_obj   = $c_class->new($c_id, %$c_cfg);
        die "Component must be instance of 'Mentat::Processor::Component'\n" unless $c_obj->isa('Mentat::Processor::Component');

        $c_type  = $c_obj->type();
        $c_name  = $c_obj->name();

        $self->{COMPONENTS}->{$c_id} = $c_obj;
        push(@{$self->{COMPONENT_MAP}->{":"}}, $c_obj);
        push(@{$self->{COMPONENT_MAP}->{"$c_type"}}, $c_obj);
        push(@{$self->{COMPONENT_MAP}->{":$c_name"}}, $c_obj);
        push(@{$self->{COMPONENT_MAP}->{"$c_type:$c_name"}}, $c_obj);

        $c_obj->register($self);
    }

    my @tmp = $self->_component_list();
    Log::Loger->info(__PACKAGE__, "[PCS] Successfully configured ".scalar(@tmp)." processor component(s)");

    return $self;
}

# _configure_links() [PROTECTED]
#
#  Usage     : $self->_configure_links();
#  Purpose   : Configure inter-component links
#  Arguments : None
#  Returns   : Mentat::Processor reference $self
#  Throws    : Dies, if not invoked on instance
#              Dies, if given invalid arguments
#  Warning   : Must be called after _configure_components() method

sub _configure_links()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my ($c_id, $c_r, $c_rid);
    my $cnt = 0;
    foreach my $c ($self->_component_list()) {
        $c_id = $c->id();
        if ($self->{LINKS}->{$c_id}) {
            foreach my $req (keys(%{$self->{LINKS}->{$c_id}})) {
                die "Component '$req' was requested in '".$c->spec()."', but no such component exists\n" unless $self->{COMPONENT_MAP}->{$req};
                $c_rid = $self->{LINKS}->{$c_id}->{$req};
                foreach my $c_r (@{$self->{COMPONENT_MAP}->{$req}}) {
                    $c->component_attached($c_r, $req, $c_rid) if $c_r->id() != $c_id;
                    $cnt++;
                }
            }
        }
    }

    Log::Loger->info(__PACKAGE__, "[PCS] Successfully configured $cnt inter-component link(s)");

    return $self;
}

# _check() [PUBLIC]
#
#  Usage     : $processor->_check();
#  Purpose   : Perform check of processor itself and all its components
#  Arguments : None
#  Returns   : Mentat::Processor reference $self
#  Throws    : Dies, if not invoked on instance
#              Dies, if error is detected
#  Comments  : Should be called after the bootstrap() method (prior to usage)
#  See Also  : bootstrap() method

sub _check()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my $tmp;

    # Check that we have working scheduler
    die "Event scheduler was not initialized\n" unless $self->{SCHEDULER};
    # Without any components processor is meaningless
    die "There are no valid components\n" unless scalar keys(%{$self->{COMPONENTS}});

    # Every event must have both emitters and handlers defined
    foreach my $event (keys(%{$self->{HANDLERS}})) {
        $tmp = $self->{HANDLERS}->{$event}->{D_E()};
        #die "Event '$event' does not have any emitters\n" unless $tmp and scalar @$tmp;
        Log::Loger->warn(__PACKAGE__, "[PCS] Event '$event' does not have any emitters") unless $tmp and scalar @$tmp;
        $tmp = $self->{HANDLERS}->{$event}->{D_H()};
        die "Event '$event' does not have any handlers\n" unless $tmp and scalar @$tmp;
    }

    # Let each component perform its own checks
    foreach my $component ($self->_component_list()) {
        $component->check();
    }

    Log::Loger->info(__PACKAGE__, "[PCS] Successfully performed operativity check");

    return $self;
}

# _daemonize(;$) [PROTECTED]
#
#  Usage     : $self->_daemonize($config);
#  Purpose   : Perform the daemonization
#  Arguments : HASH REFERENCE $config - daemonizer configurations
#  Returns   : Mentat::Processor reference $self
#  Throws    : Dies, if not invoked on instance
#

sub _daemonize(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($config,) = @_;

    # Application daemonization is optional
    if ($config) {
        my ($config, ) = @_;
        die "Daemonization configurations must be passed via HASH reference in '".CSECTION_DAEMON."' configuration section\n" unless ref $config eq 'HASH';
        die "Daemon PID file must be specified in '".CSECTION_DAEMON."' configuration section under key '".CKEY_PID_FILE."'\n" unless exists $config->{CKEY_PID_FILE()};
        die "Daemon user name must be specified in '".CSECTION_DAEMON."' configuration section under key '".CKEY_USER."'\n" unless exists $config->{CKEY_USER()};

        unless ($config->{CKEY_INFOREG()}) {
            App::Feature::Daemon->initiate($config->{CKEY_PID_FILE()}, $config->{CKEY_USER()}, $config->{CKEY_GROUP()});

            Log::Loger->info(__PACKAGE__, "[PCS] Successfully daemonized application");
        }
        else {
            Log::Loger->info(__PACKAGE__, "[PCS] Daemonization skipped, staying in foreground");
        }
    }

    return $self;
}

# _cleanup() [PROTECTED]
#
#  Usage     : $self->_cleanup();
#  Purpose   : Perform the cleanup
#  Arguments : HASH REFERENCE $config - daemonizer configurations
#  Returns   : Mentat::Processor reference $self
#  Throws    : Dies, if not invoked on instance
#

sub _cleanup()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    map { delete $self->{HANDLERS}->{$_}->{D_E()}; } keys(%{$self->{HANDLERS}});
    delete $self->{COMPONENT_MAP};
    delete $self->{LINKS};

    Log::Loger->info(__PACKAGE__, "[PCS] Successfully cleaned up after bootstrap");

    return $self;
}

#-------------------------------------------------------------------------------

# _startup(;$) [PUBLIC]
#
#  Usage     : $processor->_startup($config);
#  Purpose   : Perform startup operations
#  Arguments : None
#  Returns   : Nothing
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _startup(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($config,) = @_;

    # Startup configuration is optional
    if ($config) {
        # Deal with the initial startup events
        if ($config->{CKEY_EVENTS()}) {
            my ($emission, @eargs);
            die "Initial startup events must be passed via ARRAY reference\n" unless ref $config->{CKEY_EVENTS()} eq 'ARRAY';
            foreach my $event (@{$config->{CKEY_EVENTS()}}) {
                die "Invalid event format, expecting ARRAY reference\n" unless ref $event eq 'ARRAY';
                ($emission, @eargs) = @$event;
                die "Event emission/execution type must be passed as first parameter\n" unless $emission;
                die "Invalid event emission/execution type '$emission'\n" unless $emission =~ /emit_event(_f|_at|_pri|_pri_f|_pri_at)?|execute_event/;
                $self->$emission(@eargs);
            }
        }
    }

    # Let each component perform its own startup operations
    foreach my $component ($self->_component_list()) {
        $component->startup(@_);
    }
}

# _loop(;$) [PUBLIC]
#
#  Usage     : $processor->_loop($config);
#  Purpose   : Initiate the event loop
#  Arguments : None
#  Returns   : Nothing
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _loop(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($config,) = @_;

    # Work until the end of time
    my ($event, $next_time, $eid, $ename, $ets, @eargs);
    MAIN: until ($self->done()) {

        # Check the priority scheduler first
        ($event, $next_time) = $self->{SCHEDULER_PRI}->next_event();
        if ($event) {
            ($eid, $ename, $ets, @eargs) = @$event;
            ### Found event: $ename
            $self->execute_event($ename, @eargs);
            next;
        }
        if ($next_time) {
            ### Priority event queue is not empty, waiting for appropriate time
            next;
        }
        # Get next event from the regular scheduler and execute it
        ($event, $next_time) = $self->{SCHEDULER}->next_event();
        if ($event) {
            ($eid, $ename, $ets, @eargs) = @$event;
            ### Found event: $ename
            $self->execute_event($ename, @eargs);
        }
        else {
            ### Event queue is empty
            Log::Loger->notice(__PACKAGE__, "[PCS] Event queue is empty, done with processing");
            $self->done(1);
        }
    }
}

# _shutdown(;$) [PUBLIC]
#
#  Usage     : $processor->_shutdown($config);
#  Purpose   : Perform shutdown operations
#  Arguments : None
#  Returns   : Nothing
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _shutdown(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($config,) = @_;

    # Deal with final shutdown events (these can be only executed, event scheduler
    # is not used anymore)
    if ($config and $config->{CKEY_EVENTS()}) {
        die "Final shutdown events must be passed via ARRAY reference\n" unless ref $config->{CKEY_EVENTS()} eq 'ARRAY';
        foreach my $event (@{$config->{CKEY_EVENTS()}}) {
            die "Invalid event format, expecting ARRAY reference\n" unless ref $event eq 'ARRAY';
            $self->execute_event(@$event);
        }
    }

    # Let each component perform its own shutdown operations
    foreach my $component ($self->_component_list()) {
        $component->shutdown(@_);
    }
}

#-------------------------------------------------------------------------------

# _sequence_begin($$) [PUBLIC]
# _sequence_end($$) [PUBLIC]
#
#  Usage     : $processor->_sequence_begin($id, $title);
#              $processor->_sequence_end($id, $title);
#  Purpose   : Mark the sequence beginning/end in the log file and statistics
#  Arguments : STRING $id - sequence ID (for duration calculation)
#              STRING $title - title to be written to the log file
#  Returns   : Nothing
#  Throws    : Dies, if not invoked on instance or given invalid arguments

sub _sequence_begin($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($id, $title) = @_;

    $self->statistics->{'ts_'.$id.'_b'} = Time::HiRes::time;
    Log::Loger->notice(__PACKAGE__, '[PCS] ********** '.$title.' **********');
}
sub _sequence_end($$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($id, $title) = @_;

    $self->statistics->{'ts_'.$id.'_e'} = Time::HiRes::time;
    $self->statistics->{'ts_'.$id.'_d'} = sprintf('%.3f',$self->statistics->{'ts_'.$id.'_e'} - $self->statistics->{'ts_'.$id.'_b'});
    $title .= '...('.Value::Convertor->duration_to_string($self->statistics->{'ts_'.$id.'_d'}).')';
    Log::Loger->notice(__PACKAGE__, '[PCS] ********** '.$title.' **********');
}

#-------------------------------------------------------------------------------

sub _dump_setup()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    my $output = '';
    my @handlers;
    foreach my $event (sort keys(%{$self->{HANDLERS}})) {
        $output .= "    *  $event ==> ";
        @handlers = ();
        foreach my $handler (@{$self->{HANDLERS}->{$event}->{D_H()}}) {
            push (@handlers, "$handler->[0]:$handler->[1]");
        }
        $output .= join(' ---> ', @handlers) . "\n";
    }
    $output =~ s/\n$//;
    Log::Loger->info(__PACKAGE__, "[PCS] Processor component setup:\n$output");
}

=pod

=back

=cut

1;
