package Mentat::Processor::Component;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Processor::Component - Short module description

=head1 SYNOPSIS

 use Mentat::Processor::Component;

=head1 DESCRIPTION

Extensive module description

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.

=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

#use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Mentat::Core qw(:c_general :c_processor);

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#
use constant BASE_NAME => 'component';
use constant BASE_TYPE => 'component';

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.1';
    $DEVEL   = 0;
}


#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

# Operator overloading
use overload '""' => 'to_string';

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new [PUBLIC,STATIC]

  Usage     : my $instance = Mentat::Processor::Component->new()
  Purpose   : Create and return reference to the new instance
  Arguments : MIXED $id - component unique identifier [MANDATORY]
              HASH %config - component configuration
  Returns   : Mentat::Processor::Component reference on success, dies on failure
  Throws    : Dies, if not invoked on class
  Comments  : Internally uses _init() method, subclasses should never overload this method.
  See Also  : _init() method

=cut

sub new
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    my ($id, %config) = @_;
    confess "Component's unique identifier must be given as first argument" unless $id;

    ### Creating Mentat.Processor.Component instance...
    my $self = bless ({}, $class);
    $self->{ID} = $id;
    $self->{STATISTICS} = {};
    return $self->_init(%config);
}

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if invoked on class

=cut

sub DESTROY()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    ### Destroying Mentat.Processor.Component instance...
    # This is vital to break the reference cycles
    delete $self->{PROCESSOR};
}

#-------------------------------------------------------------------------------

=item type() [PUBLIC]

  Usage     : my $name = $component->type();
  Purpose   : Get the type of the component (first chunk after Mentat::Processor::Component)
  Arguments : None
  Returns   : STRING $name lowercase component name
  Throws    : Dies, if not invoked on instance

=cut

sub type()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    # Lazy-determine the type of the component (first chunk after Mentat::Processor::Component)
    unless ($self->{TYPE}) {
        my $name = blessed($self);
        confess "Can't automatically determine the type of the component" unless $name =~ /Mentat::Processor::Component/;

        $name =~ s/Mentat::Processor::Component//; $name =~ s/^:://;
        my @tmp = split(/:+/, $name);
        if (@tmp) {
            $self->{TYPE} = lc($tmp[0]);
        }
        else {
            $self->{TYPE} = BASE_TYPE;
        }
    }
    return $self->{TYPE};
}

=item name() [PUBLIC]

  Usage     : my $name = $component->name();
  Purpose   : Get the name of the component (last chunk of the class name)
  Arguments : None
  Returns   : STRING $name lowercase component name
  Throws    : Dies, if not invoked on instance

=cut

sub name()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    # Lazy-determine the name of the component (last part of the class name)
    unless ($self->{NAME}) {
        my $name = blessed($self);
        confess "Can't automatically determine the name of the component" unless $name =~ /Mentat::Processor::Component/;

        $name =~ s/Mentat::Processor::Component//; $name =~ s/^:://;
        my @tmp = split(/:+/, $name);
        if (@tmp) {
            $self->{NAME} = lc($tmp[$#tmp]);
        }
        else {
            $self->{NAME} = BASE_NAME;
        }
    }
    return $self->{NAME};
}

=item spec() [PUBLIC]

  Usage     : my $name = $component->spec();
  Purpose   : Get the specification of the component ("$type:$name")
  Arguments : None
  Returns   : STRING $spec lowercase component specification
  Throws    : Dies, if not invoked on instance

=cut

sub spec()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);

    return $self->type().':'.$self->name();
}

=item id() [PUBLIC]

  Usage     : my $id = $component->id();
  Purpose   : Get the unique identifier of the component
  Arguments : None
  Returns   : STRING $id lowercase component unique ID
  Throws    : Dies, if not invoked on instance

=cut

sub id()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->{ID};
}

=item to_string() [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub to_string()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return join(':',$self->spec,$self->id);
}

=item statistics() [PUBLIC]

  Usage     : my $statistics = $processorstatistics();
  Purpose   : Get component statistics
  Arguments : None
  Returns   : HASH REFERENCE $statistics with various keys
  Throws    : Dies, if not invoked on class
  Comments  : Mainly for debugging, development and optimization purposes

=cut

sub statistics()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->{STATISTICS};
}

#-------------------------------------------------------------------------------

=item processor(;$) [PUBLIC]

  Usage     : Getter: my $processor = $instance->processor();
              Setter: $instance->processor($processor);
  Purpose   : Set/Retrieve the Mentat::Processor reference
  Arguments : Mentat::Processor $processor - main controller [OPTIONAL]
  Returns   : Mentat::Processor $processor
  Throws    : Dies, if not invoked on instance or given invalid arguments

=cut

sub processor(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($processor,) = @_;

    if ($processor) {
        #confess "Expecting instance of Mentat::Processor as argument" unless blessed($processor) and $processor->isa('Mentat::Processor');
        $self->{PROCESSOR} = $processor;
    }
    return $self->{PROCESSOR};
}

#-------------------------------------------------------------------------------

=item register($) [PUBLIC]

  Usage     : my $error = $instance->register($processor);
  Purpose   : Register the component into given Mentat::Processor main controller
  Arguments : Mentat::Processor $processor - main controller [MANDATORY]
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance or given invalid arguments
  Comment   : Template method, requires implementation of _get_events_emitted()
              and _get_events_handled() hook methods in descendant classes.
  Comment   : Called during Mentat::Processor`s bootstrap phase
  See Also  : Mentat::Processor class for register_event() and register_event_handler()
              methods

=cut

sub register($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($processor,) = @_;
    confess "Expecting instance of Mentat::Processor as first argument" unless blessed($processor) and $processor->isa('Mentat::Processor');
    my ($items, $event, $handler);

    ### Register the processor main controller
    $self->processor($processor);

    ### Register the list of emited events into main controller
    $items = $self->_get_events_emitted();
    for my $item (@$items) { ### Registering events [===|    ] % done
        $self->processor->register_event($self, $item);
    }

    ### Register the list of handled events into main controller
    $items = $self->_get_events_handled();
    for my $item (@$items) { ### Registering handlers [===|    ] % done
        $self->processor->register_event_handler($self, @$item);
    }

    ### Register the list of required components into main controller
    $items = $self->_get_components_linked();
    for my $item (@$items) { ### Registering linked components [===|    ] % done
        $self->processor->register_linked_component($self, @$item);
    }
}

=item component_attached($$;$) [PUBLIC]

  Usage     : $instance->component_attached($requested_component, $specification, $id);
  Purpose   : Allow the component to react to attachment of other components
  Arguments : Mentat::Processor::Component $component - component, that was attached
              STRING $specification - specification of the attached component
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance or given invalid arguments
              Dies, in case of some problems with other component
  Comment   : Called during Mentat::Processor`s bootstrap phase

=cut

sub component_attached($$;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($component, $specification, $id) = @_;
    $id = 'component' if not $id or "$id" eq 1;

    die "Component with ID '$id' was already attached" if exists $self->{COMPONENTS}->{$id};
    $self->{COMPONENTS}->{$id} = $component;
}

=item check() [PUBLIC]

  Usage     : $instance->check();
  Purpose   : Callback to allow the component perform last check before startup
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance or given invalid arguments
              Dies in case of some problems with component configuration

=cut

sub check()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
}

#-------------------------------------------------------------------------------

=item startup(;$) [PUBLIC]

  Usage     : $instance->startup();
  Purpose   : Callback to allow the component perform arbitrary startup operations
  Arguments : HASH REFERENCE $config - startup configurations [OPTIONAL]
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance or on error

=cut

sub startup(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($config,) = @_;
    confess "Component startup configurations must be passed via HASH reference" if $config and ref $config ne 'HASH';
}

=item shutdown(;$) [PUBLIC]

  Usage     : $instance->shutdown();
  Purpose   : Callback to allow the component perform arbitrary shutdown operations
  Arguments : HASH REFERENCE $config - shutdown configurations [OPTIONAL]
  Returns   : Nothing
  Throws    : Dies, if not invoked on instance or on error

=cut

sub shutdown(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($config,) = @_;
    confess "Component shutdown configurations must be passed via HASH reference" if $config and ref $config ne 'HASH';
}

#-------------------------------------------------------------------------------

=item emit_event($;@) [PUBLIC]

  Usage     : my $id = $instance->emit_event($event, @args);
  Purpose   : Emit new event and schedule it at the end of event queue
  Arguments : STRING $event - event name [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : MIXED $id unique event identifier
  Throws    : Dies, if not invoked on class or given invalid arguments

=cut

sub emit_event($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->processor->emit_event(@_);
}

=item emit_event_f($;@) [PUBLIC]

  Usage     : my $id = $instance->emit_event_f($event, @args);
  Purpose   : Emit new event and schedule it at the beginning of event queue
  Arguments : STRING $event - event name [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : MIXED $id unique event identifier
  Throws    : Dies, if not invoked on class or given invalid arguments

=cut

sub emit_event_f($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->processor->emit_event_f(@_);
}

=item emit_event_at($$;@) [PUBLIC]

  Usage     : my $id = $instance->emit_event_at($event, @args);
  Purpose   : Emit new event and schedule it at specific time
  Arguments : STRING $event - event name [MANDATORY]
              INTEGER $timestamp - unix timestamp [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : MIXED $id unique event identifier
  Throws    : Dies, if not invoked on class or given invalid arguments

=cut

sub emit_event_at($$;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->processor->emit_event_at(@_);
}

=item emit_event_pri($;@) [PUBLIC]

  Usage     : my $id = $instance->emit_event_pri($event, @args);
  Purpose   : Emit new priority event and schedule it at the end of event queue
  Arguments : STRING $event - event name [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : MIXED $id unique event identifier
  Throws    : Dies, if not invoked on class or given invalid arguments

=cut

sub emit_event_pri($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->processor->emit_event_pri(@_);
}

=item emit_event_pri_f($;@) [PUBLIC]

  Usage     : my $id = $instance->emit_event_pri_f($event, @args);
  Purpose   : Emit new priority event and schedule it at the beginning of event queue
  Arguments : STRING $event - event name [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : MIXED $id unique event identifier
  Throws    : Dies, if not invoked on class or given invalid arguments

=cut

sub emit_event_pri_f($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->processor->emit_event_pri_f(@_);
}

=item emit_event_pri_at($$;@) [PUBLIC]

  Usage     : my $id = $instance->emit_event_pri_at($event, @args);
  Purpose   : Emit new priority event and schedule it at specific time
  Arguments : STRING $event - event name [MANDATORY]
              INTEGER $timestamp - unix timestamp [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : MIXED $id unique event identifier
  Throws    : Dies, if not invoked on class or given invalid arguments

=cut

sub emit_event_pri_at($$;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->processor->emit_event_pri_at(@_);
}

=item execute_event($;@) [PUBLIC]

  Usage     : my $id = $instance->execute_event($event, @args);
  Purpose   : Bypass event emission and immediatelly start to execute chain of the
              given event
  Arguments : STRING $event - event name [MANDATORY]
              MIXED @arguments - event arguments [OPTIONAL]
  Returns   : BOOL $result - TRUE on SUCCESS, FALSE on FAILURE
  Throws    : Dies, if not invoked on class or given invalid arguments

=cut

sub execute_event($;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->processor->execute_event(@_);
}

#-------------------------------------------------------------------------------

=item done(;$) [PUBLIC]

  Usage     : $instance->done(1); # Set the DONE flag
  Purpose   : After setting up this flag, processor will gracefully terminate
  Arguments : BOOL $done_flag - any perlish TRUE value will turn the flag ON
  Returns   : INTEGER 1, if the processor will stop, INTEGER 0 otherwise
  Throws    : Dies, if not invoked on instance
  Comments  : Once the done flag is set, it can`t be cleared
  See Also  : Mentat::Processor::done() and Mentat::Processor::run() methods

=cut

sub done(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->processor->done(@_);
}

#-------------------------------------------------------------------------------

sub pause()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->processor->pause(@_);
}
sub resume()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->processor->resume(@_);
}

#-------------------------------------------------------------------------------

sub state($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($new_state,) = @_;

    $self->{STATE} = "$new_state" if defined $new_state;
    return $self->{STATE};
}
sub is_paused()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->{STATE} eq STATE_PAUSED;
}

#-------------------------------------------------------------------------------

=item wait($) [PUBLIC]

  Usage     : $component->wait($time);
  Purpose   : Wait for given amount of time (sleep)
  Arguments : FLOAT $time - how long to wait
  Returns   : FLOAT $slept how long actually waited
  Throws    : Dies, if not invoked on instance

=cut

sub wait($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->processor->wait(@_);
}

#-------------------------------------------------------------------------------

=item status_msg($) [PUBLIC]

  Usage     : $component->status_msg($message);
  Purpose   : Setup the status message, which will be sent to the user after processing is done
  Arguments : STRING $message - status message [MANDATORY]
  Returns   : Nohing
  Throws    : Dies, if not invoked on instance

=cut

sub status_msg($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->processor->status_msg(@_);
}

=item runlog_set($) [PUBLIC]

  Usage     : $component->runlog_set($runlog);
  Purpose   : Setup the processing runlog, it will be stored automatically
  Arguments :
  Returns   : Nohing
  Throws    : Dies, if not invoked on instance

=cut

sub runlog_set($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->processor->runlog_set(@_);
}

=item runlog_get($) [PUBLIC]

  Usage     : $component->runlog_get();
  Purpose   : Retrieve the processing runlog
  Arguments :
  Returns   : Nohing
  Throws    : Dies, if not invoked on instance

=cut

sub runlog_get()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->processor->runlog_get();
}

#-------------------------------------------------------------------------------

=item component($;$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    :
  Comments  :
  See Also  :

=cut

sub component($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($comp, $value) = @_;

    $self->{COMPONENTS}->{$comp} = $value if defined $value;

    return $self->{COMPONENTS}->{$comp};
}

=item start_time(;$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub start_time(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->processor->start_time(@_);
}

=item duration(;$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    : Dies, if not invoked on instance

=cut

sub duration(;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return $self->processor->duration(@_);
}

#=item AUTOLOAD [PUBLIC]

  #Usage     : my $value = $component->value;
  #Purpose   : Direct access to the attributes via the autogenerated access methods
  #Arguments :
  #Returns   :
  #Throws    :
  #Comments  :
  #See Also  :

#=cut

#sub AUTOLOAD {
    #my $self = shift;
    #confess "Instance method not invoked on object instance" unless blessed($self);
    #my $c = our $AUTOLOAD;

    ## Perform some sanity checks first
    #$c =~ s/.*:://;
    #return if $c =~ /[A-Z]+/;
    #return unless $c =~ /[_a-zA-Z][_a-zA-Z0-9]*/;

    ## We do not want any missspelings here, so croak on error
    #confess "Requested invalid linked component '$c'" unless exists($self->{COMPONENTS}->{$c});

    ## Try to retrieve and return the attribute
    #return $self->component($c);
#}

#-------------------------------------------------------------------------------
# Private interface
#-------------------------------------------------------------------------------

# _init(%) [PROTECTED]
#
#  Usage     : return $self->_init(@_);
#  Purpose   : Initialize the new Mentat::Processor::Component instance
#  Arguments : HASH %config - component configurations
#  Returns   : Mentat::Processor::Component reference
#  Throws    : Dies, if not invoked on instance
#  Comments  : This method must never be called directly, it is supposed to be overloaded by subclasses.

sub _init(%)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my %config = @_;

    $self->{COMPONENTS} = {};
    $self->{STATE} = STATE_NORMAL;

    return $self;
}

# _join_lists(@) [PROTECTED]
#
#  Usage     : my $list = $self->_join_lists($list1, $list2);
#  Purpose   : Join given lists into one with unique items
#  Arguments : LIST OF ARRAY REFERENCES @lists - lists to join [MANDATORY]
#  Returns   : ARRAY REFERENCE $list
#  Throws    : Dies, if not invoked on instance

sub _join_lists(@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my @lists = @_;
    my %result = ();

    # Use indirrect list join via hash to make the items unique
    foreach my $list (@lists) {
        foreach my $i (@$list) {
            $result{$i}++;
        }
    }
    my @list = keys(%result);
    return \@list;
}

# _join_array_lists(@) [PROTECTED]
#
#  Usage     : my $list = $self->_join_array_lists($list1, $list2);
#  Purpose   : Join given event handler lists into one with unique items
#  Arguments : LIST OF ARRAY REFERENCES @lists - lists to join [MANDATORY]
#  Returns   : ARRAY REFERENCE $list
#  Throws    : Dies, if not invoked on instance

sub _join_array_lists(@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my @lists = @_;
    my %result = ();

    # Use indirrect list join via hash to make the items unique
    my $tmp;
    foreach my $list (@lists) {
        foreach my $i (@$list) {
            $tmp = $i->[1] || $i->[0];
            $result{$i->[0]}->{$tmp}++;
        }
    }

    my @list = ();
    foreach my $k (keys(%result)) {
        foreach my $j (keys(%{$result{$k}})) {
            push(@list, [$k,$j]);
        }
    }
    return \@list;
}

#-------------------------------------------------------------------------------

# _get_events_emitted(;@) [PROTECTED]
#
#  Usage     : Used as a hook method from template method register()
#  Purpose   : Return list of events emited by this component
#  Arguments : LIST OF ARRAY REFERENCES @lists - event lists from parent classess [OPTIONAL]
#  Returns   : ARRAY REFERENCE $events containing event names as STRINGS
#  Throws    : Dies, if not invoked on instance

sub _get_events_emitted(;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my @lists = @_;
    return $self->_join_lists([], @lists);
}

# _get_events_handled(;@) [PROTECTED]
#
#  Usage     : Used as a hook method from template method register()
#  Purpose   : Return list of events and appropriate handlers for this component
#  Arguments : LIST OF ARRAY REFERENCES @lists - event handler lists from parent classess [OPTIONAL]
#  Returns   : ARRAY REFERENCE $handlers with internal structure [$event, $handler]
#                  STRING $event   - event name
#                  STRING $handler - event handler name
#  Throws    : Dies, if not invoked on instance

sub _get_events_handled(;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my @lists = @_;
    return $self->_join_array_lists([], @lists);
}

# _get_components_linked(;@) [PROTECTED]
#
#  Usage     : Used as a hook method from template method register()
#  Purpose   : Return list of required components
#  Arguments : LIST OF ARRAY REFERENCES @lists - linked component lists from parent classess [OPTIONAL]
#  Returns   : ARRAY REFERENCE $components containing STRINGS with following format:
#                  'type'      - request for component of given type and any name
#                  ':name'     - request for component of any type but given name
#                  'type:name' - request for component of given both type and name
#  Throws    : Dies, if not invoked on instance

sub _get_components_linked(;@)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my @lists = @_;
    return $self->_join_array_lists([], @lists);
}

=pod

=back

=cut

1;
