package Mentat::Processor::Message;
use strict;
use warnings;

#*******************************************************************************
# DOCUMENTATION SECTION
#*******************************************************************************

=head1 NAME

Mentat::Processor::Message - Mentat processor message class

=head1 SYNOPSIS

 use Mentat::Processor::Message;

=head1 DESCRIPTION

=head1 AUTHOR

 Jan Mach
 Cesnet, z.s.p.o
 jan.mach@cesnet.cz
 http://www.cesnet.cz

=head1 COPYRIGHT

This program is free software; you can redistribute
it and/or modify it under the same terms as Perl itself.

The full text of the license can be found in the
LICENSE file included with this module.


=head1 SEE ALSO

perl(1).

=head1 FUNCTION REFERENCE

=over 4

=cut

#*******************************************************************************
# LIBRARY LOADING SECTION
#*******************************************************************************

#-- Perl core modules ---------------------------------------------------------#
use Carp;
use Scalar::Util 'blessed';

use Data::Dumper; #-+-> DEVEL ONLY <-+-#
#use Smart::Comments; #-+-> DEVEL ONLY <-+-#

#-- Perl CPAN modules ---------------------------------------------------------#

#-- Custom application modules ------------------------------------------------#
use Log::Loger;

#*******************************************************************************
# CONSTANTS AND GLOBAL VARIABLES DEFINITION SECTION
#*******************************************************************************

#-- Constants -----------------------------------------------------------------#

#-- Static public class variables (our) ---------------------------------------#

#-- Static protected class variables (my) -------------------------------------#

#*******************************************************************************
# INITIALIZATION AND CLEANUP SECTION
#*******************************************************************************

#-- Module initializations ----------------------------------------------------#
BEGIN {
    use vars qw($VERSION $DEVEL);
    $VERSION = '0.2';
    $DEVEL   = 0;
}

#-- Module clean-up code (global destructor) ----------------------------------#
END {

}

# Operator overloading
use overload '""' => 'to_string';

#*******************************************************************************
# IMPLEMENTATION SECTION
#*******************************************************************************

=item new [PUBLIC,STATIC]

  Usage     : my $instance = Mentat::Processor::Message->new()
  Purpose   : Create and return reference to the new instance
  Arguments : Same as the the _init() method
  Returns   : Mentat::Processor::Message reference on success, dies on failure
  Throws    : Dies, if not invoked on class
  Comments  : Internally uses _init() method
  See Also  : _init() method

=cut

sub new
{
    my $class = shift;
    confess "Class method not invoked on class" if blessed($class);
    ### Creating Mentat.Processor.Message instance...

    my $self = bless ({}, $class);
    $self->{ATTRIBUTES} = {};
    return $self;
}

=item DESTROY() [PUBLIC]

  Usage     : This method must not be used directly
  Purpose   : Class destructor
  Arguments : None
  Returns   : Nothing
  Throws    : Dies, if invoked on class

=cut

sub DESTROY()
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    ### Destroying Mentat.Processor.Message instance...
}

#-------------------------------------------------------------------------------

=item has_attribute($) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    :

=cut

sub has_attribute($)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($attr,) = @_;

    return exists($self->{ATTRIBUTES}->{$attr});
}

=item attribute($;$) [PUBLIC]

  Usage     :
  Purpose   :
  Arguments :
  Returns   :
  Throws    :

=cut

sub attribute($;$)
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my ($attr, $value) = @_;

    $self->{ATTRIBUTES}->{$attr} = $value if defined $value;

    return $self->{ATTRIBUTES}->{$attr};
}

=item AUTOLOAD [PUBLIC]

  Usage     : my $value = $request->value;
  Purpose   : Direct access to the attributes via the autogenerated access methods
  Arguments :
  Returns   :
  Throws    :

=cut

sub AUTOLOAD {
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    my $attr = our $AUTOLOAD;

    # Perform some sanity checks first
    $attr =~ s/.*:://;
    return if $attr =~ /[A-Z]+/;
    return unless $attr =~ /[_a-zA-Z][_a-zA-Z0-9]*/;

    # We do not want any missspelings here, so croak on error
    confess "Requested invalid attribute '$attr'" unless exists($self->{ATTRIBUTES}->{$attr});

    # Try to retrieve and return the attribute
    return $self->attribute($attr);
}

#-------------------------------------------------------------------------------

=item to_string [PUBLIC]

  Usage     :
  Purpose   : Convert the request to string (for debugging purposes)
  Arguments :
  Returns   :
  Throws    :

=cut

sub to_string
{
    my $self = shift;
    confess "Instance method not invoked on object instance" unless blessed($self);
    return Dumper($self);
}

=pod

=back

=cut

1;
